#linux.s - Linux-specific word definitions.

	.include "forth.inc"

	.set stdin, 0
	.set stdout, 1
	.set stderr, 2

	.set O_RDWR, 2
	.set SEEK_SET, 0

	.set NR_exit, 1
	.set NR_read, 3
	.set NR_write, 4
	.set NR_open, 5
	.set NR_close, 6
	.set NR_lseek, 19
	.set NR_ioctl, 54


	space

argc:	.int 0
cur_arg: .int 0
cur_fd: .int 0
file_buf: .int 0
#Terminal I/O Setting structures.
ti:	.fill 44,1,0
ti_old:	.fill 44,1,0


	code

	.global f_emit
f_emit:
	li r5,1		#1 byte
	stw tos,-4(ds)	#store cell on top of stack,
	addi r4,ds,-1	#and get address of LSB
	li r3,stdout	#stdout
	li r0,NR_write	#write syscall #
	sc
	drop
	blr

	.global f_type
f_type:
	li r3,stdout	#stdout
	nip r4		#address
	mr r5,tos	#length
	drop
	li r0,NR_write	#write syscall #
	sc
	blr

	.global err_print
	#r4 = address, r5 = length
err_print:
	li r3,stderr
	li r0,NR_write
	sc
	blr

open_next:
	#more arguments to process?
	lwzi r4,r3,argc
	cmpwi r4,0
	bgt 1f
	lwzi r4,r3,cur_fd
	cmpwi r4,stdin
	bne 0f
	#no more input, exit.
	li r3,INPUT_END_CODE
	b error
0:	li r4,stdin
	stw r4,cur_fd@l(r3)
	mr r3,r4
	blr
1:	#decrement argc
	addi r4,r4,-1
	stw r4,argc@l(r3)
	#increment arg pointer
	lwzi r6,r5,cur_arg
	addi r6,r6,4
	stw r6,cur_arg@l(r5)
	#open file
	lwz r3,-4(r6)
	li r4,O_RDWR
	li r0,NR_open
	sc
	#store file descriptor
	stwi r4,r3,cur_fd
	blr

close_file:
	lwzi r3,r3,cur_fd
	li r0,NR_close
	sc
	blr

	.global f_refill
f_refill:
	prolog
	lwzi r3,r3,cur_fd
	cmpwi r3,-1
0:	beql open_next
	lwzi r4,r4,file_buf
	li r5,1024
	li r0,NR_read
	sc
	cmpwi r3,0
	bne 1f
	#end of file, close it, and start the next one.
	bl close_file
	crset eq #set eq flag, so open_next will get called.
	b 0b
	#done reading file, set up input buffer and return.
1:	liw r4,input
	#store address and length of buffer
	lwzi r5,r5,file_buf
	stw r5,in_buf(r4)
	stw r3,in_len(r4)
	#reset input index
	li r3,0
	stw r3,in_cur(r4)
	epilog

	.global f_bye, exit
f_bye:
	li r3,OK_CODE
exit:
	stwu r3,-4(rs)
	bl restore_term
	lwz r3,0(rs)
	addi rs,rs,4
	li r0,NR_exit
	sc

	.global f_err
f_err:
	li r3,ERR_CODE
	b error

save_term: #save old terminal I/O settings
	li r3,stdin
	liwl r4,0x402c7413
	liw r5,ti_old
	li r0,NR_ioctl
	sc
	blr

set_term: #modify terminal I/O settings
	#get settings,
	li r3,stdin
	liwl r4,0x402c7413
	liw r5,ti
	li r0,NR_ioctl
	sc
	#modify them,
	liw r4,ti
	lwz r3,12(r5) #modify lflags
	li r4,~0x180 #clear ISIG, ICANON
	and r3,r3,r4
	ori r3,r3,0x8 #set ECHO
	stw r3,12(r5)
	#and write them back.
	li r3,stdin
	liwl r4,0x802c7414
	li r0,NR_ioctl
	sc
	blr

restore_term: #restore old terminal I/O settings
	li r3,stdin
	liwl r4,0x802c7414
	liw r4,ti_old
	li r0,NR_ioctl
	sc
	blr

	.global _start
_start:
	bl init
	#set up argument list.
	liw r3,argc
	lwz r4,0(rs)
	addi r4,r4,-1
	addi r5,rs,8
	stw r4,0(r3)
	stw r5,4(r3)
	#set up file input.
	stwi here,r3,file_buf
	addi here,here,1024
	li r3,-1
	stwi r3,r4,cur_fd
	#set terminal settings
	bl save_term
	#bl set_term #clear ISIG, ICANON, set ECHO
	#start forth.
	b run
