#main.s - basic forth.  start here, then rebuild from forth source.

	.include "forth.inc"

	data

	.global input
input: .int 0, 0, 0  #addr, length, cur-offset.

	cdata

nf_str:	.byte 0f-(.+1)
	.byte '"'
	.byte " not found"
	.byte 10
0:


	code

	.global init
init:
	#initialize registers
	liw toc,heap		#ToC Register
	liw ds,ds_base		#data stack pointer
	mr here,toc             #current heap pointer
	#initialize global variables
	li r3,first_free_var	#var_here
	stw r3,var_here(toc)
	li r4,INTERPRET		#start off in interpret state
	stw r4,state(toc)
	stw here,flushed_to(toc)
	li r3,10		#start off in decimal
	stw r3,base(toc)
	#initialize dictionaries
	liw r3,forth_mem	#forth dictionary
	stw r3,forth_dict(toc)
	stw r3,forth_free(toc)
	liw r4,compiler_mem	#compiler dictionary
	stw r4,compiler_dict(toc)
	stw r4,compiler_free(toc)
	addi r3,toc,forth_dict	#current dictionary (one we add new names to)
	stw r3,current_dict(toc)
copy_primitives:
	#forth primitives
	liw r3,forth_dict_start
	addi r3,r3,-4
	lwz r4,forth_dict(toc)
	addi r4,r4,-4
	li r5,forth_dict_length@l
	srwi r5,r5,2
	mtctr r5
0:	lwzu r5,4(r3)
	stwu r5,4(r4)
	bdnz 0b
	addi r4,r4,4
	stw r4,forth_free(toc)
	#compiler primitives
	liw r3,compiler_dict_start
	addi r3,r3,-4
	lwz r4,compiler_dict(toc)
	addi r4,r4,-4
	li r5,compiler_dict_length@l
	srwi r5,r5,2
	mtctr r5
0:	lwzu r5,4(r3)
	stwu r5,4(r4)
	bdnz 0b
	addi r4,r4,4
	stw r4,compiler_free(toc)
	#set up input to come from built-in source.
	liw r3,input
	liw r4,src_start
	li r5,0
	lwzi r6,r6,src_len
	stw r4,in_buf(r3)
	stw r5,in_cur(r3)
	stw r6,in_len(r3)
	blr

	.global word_not_found
word_not_found:
	dup
	li tos,'"'
	bl f_emit
	bl f_type	#print out the word
	liw r4,nf_str
	lbz r5,0(r4)
	addi r4,r4,1
	bl err_print
	_epilog
	b run

interpret:
	prolog
	bl f_ffind
	cmpwi tos,0
	drop
	beq 1f
0:	bl f_execute
	b 9f
1:	bl f_to_num_q
	cmpwi tos,0
	drop
	beq word_not_found
	#don't need to do anything -- number is already on stack
9:	epilog

compile:
	prolog
	bl f_cfind
	cmpwi tos,0
	drop
0:	beq 1f
	bl f_execute
	b 9f
1:	bl f_ffind
	cmpwi tos,0
	drop
	beq 2f
	bl f_compile
	b 9f
2:	bl f_to_num_q
	cmpwi tos,0
	drop
	beq word_not_found
	bl m_literal
9:	epilog

postpone:
	prolog
	bl f_cfind
	cmpwi tos,0
	beq word_not_found
	drop
	li r3,COMPILE	#switch back to compile state
	stw r3,state(toc)
	bl f_compile
	epilog

	.global run
run:
0:	bl f_word
	lwz r3,state(toc)
	cmpwi r3,INTERPRET
	bne 1f
	bl interpret
	b 0b
1:	cmpwi r3,COMPILE
	bne 1f
	bl compile
	b 0b
1:	cmpwi r3,POSTPONE
	bne 1f
	bl postpone
	b 0b
1:	#invalid state, exit with error code.
	li r3,BAD_STATE_CODE
	b error
